"""Connection class marshals data over UDP to daemon."""
import json
from socket import AF_INET, SOCK_DGRAM, SocketType, socket

from xray_ot.encoder import default_encoder


class XRayConnection(object):
    """
    Instances of _Connection are used to communicate with the X-Ray daemon via UDP.
    """

    def __init__(self, collector_address: (str, int)):
        self._collector_address: str = collector_address
        self._socket: SocketType = socket(AF_INET, SOCK_DGRAM)

    @staticmethod
    def to_payload(msg) -> bytes:
        return (
            '{"format": "json", "version": 1}\n'
            + json.dumps(msg, default=default_encoder)
        ).encode("utf8")

    def report(self, msg) -> None:
        """Report to the daemon."""
        for m in msg:
            self._socket.sendto(self.to_payload(m), self._collector_address)



"""Constants"""

# Runtime constants
FLUSH_THREAD_NAME = "Flush Thread"
FLUSH_PERIOD_SECS = 2.5
DEFAULT_MAX_SPAN_RECORDS = 1000

# Reserved Span keys
PARENT_SPAN_GUID = "parent_span_guid"

# Log Keywords
PAYLOAD = "payload"
SPAN_GUID = "span_guid"

# Log Levels
INFO_LOG = "I"
WARN_LOG = "W"
ERROR_LOG = "E"
FATAL_LOG = "F"

# JSON pickle settings
JSON_FAIL = "<Invalid Payload"
JSON_MAX_DEPTH = 32
JSON_UNPICKLABLE = True
JSON_WARNING = True

# utils constants
SECONDS_TO_MICRO = 1000000

# Recorder constants
MAX_LOG_MEMORY = 1024
MAX_LOG_LEN = 984
JOIN_ID_TAG_PREFIX = "join:"



from xray_ot.settings import RecordingSettings
from xray_ot.translator.ot_types import T_TAGS, T_XRAY_TAGS


def http(settings: RecordingSettings, tags: T_TAGS) -> T_XRAY_TAGS:
    request = {}
    response = {}

    for tag_name, tag_value in tags.items():
        obj = response

        if tag_name in settings.http_request_annotations:
            obj = request
            tag_name = settings.http_request_annotations[tag_name]

        elif tag_name in settings.http_response_annotations:
            tag_name = settings.http_response_annotations[tag_name]

        obj[tag_name] = tag_value

    return {
        "http": {
            "request": request,
            "response": response,
        }
    }



from xray_ot.settings import RecordingSettings
from xray_ot.translator.ot_types import T_TAGS, T_XRAY_TAGS


def db(settings: RecordingSettings, tags: T_TAGS) -> T_XRAY_TAGS:
    return {
        "sql": {
            (settings.db_annotations.get(tag_name, tag_name)): tag_value
            for tag_name, tag_value in tags.items()
        }
    }



from __future__ import absolute_import

import struct
from typing import TypeVar

from basictracer.context import SpanContext
from basictracer.propagator import Propagator
from opentracing import InvalidCarrierException

T_BUFFER = TypeVar("T_BUFFER", bytearray, bytes, memoryview)


class BinaryPropagator(Propagator):
    """A BasicTracer Propagator for Format.BINARY."""

    def inject(self, span_context: SpanContext, carrier: bytearray) -> None:
        if type(carrier) is not bytearray:
            raise InvalidCarrierException()
        _wstr(carrier, span_context.trace_id)
        _wstr(carrier, span_context.span_id)
        carrier.extend(struct.pack("?", span_context.sampled))
        carrier.extend(
            struct.pack(
                ">I",
                len(span_context.baggage) if span_context.baggage is not None else 0,
            )
        )
        if span_context.baggage is not None:
            for key in span_context.baggage:
                _wstr(carrier, key)
                _wstr(carrier, span_context.baggage[key])

    def extract(self, carrier: bytearray) -> SpanContext:
        if type(carrier) is not bytearray:
            raise InvalidCarrierException()
        m = memoryview(carrier)
        m, trace_id = _rstr(m)
        m, span_id = _rstr(m)
        sampled = struct.unpack("?", m[:1])[0]
        m = m[1:]
        bl = struct.unpack(">I", m[:4])[0]
        m = m[4:]
        baggage = {}
        for _ in range(bl):
            m, k = _rstr(m)
            m, v = _rstr(m)
            baggage[k] = v

        return SpanContext(
            span_id=span_id, trace_id=trace_id, baggage=baggage, sampled=sampled
        )


def _wstr(buf: T_BUFFER, s: str) -> None:
    buf.extend(struct.pack(">I", len(s)))
    buf.extend(s.encode("utf8"))


def _rstr(buf: T_BUFFER) -> (T_BUFFER, str):
    l = struct.unpack(">I", buf[:4])[0]
    s = bytes(buf[4 : 4 + l]).decode("utf8")
    return buf[4 + l :], s



"""
AWS X-Ray backed implementation of the python OpenTracing API.

http://opentracing.io
https://github.com/opentracing/api-python

See the API definition for comments.
"""
from __future__ import absolute_import

import binascii
import os
import time

from basictracer import BasicTracer
from opentracing import Format

from .binary_propagator import BinaryPropagator
from xray_ot.recorder import Recorder
from .rand import generate_64bit_id_as_hex
from .text_propagator import TextPropagator


def Tracer(**kwargs) -> "_XRayTracer":
    """Instantiates the AWS X-Ray OpenTracing implementation.

    :param str component_name: the human-readable identity of the instrumented
        process. I.e., if one drew a block diagram of the distributed system,
        the component_name would be the name inside the box that includes this
        process.
    :param str collector_host: X-Ray daemon hostname
    :param int collector_port: X-Ray daemon port
    :param dict tags: a string->string dict of tags for the Tracer itself (as
        opposed to the Spans it records)
    :param int max_span_records: Maximum number of spans records to buffer
    :param int periodic_flush_seconds: seconds between periodic background
        flushes, or 0 to disable background flushes entirely.
    :param int verbosity: verbosity for (debug) logging, all via logging.info().
        0 (default): log nothing
        1: log transient problems
        2: log all of the above, along with payloads sent over the wire
    """
    return _XRayTracer(Recorder(**kwargs))


class _XRayTracer(BasicTracer):
    def __init__(self, recorder):
        """Initialize the X-Ray Tracer, deferring to BasicTracer."""
        super(_XRayTracer, self).__init__(recorder)
        self.register_propagator(Format.TEXT_MAP, TextPropagator())
        self.register_propagator(Format.HTTP_HEADERS, TextPropagator())
        self.register_propagator(Format.BINARY, BinaryPropagator())

    def flush(self) -> None:
        """Force a flush of buffered Span data to the X-Ray daemon."""
        self.recorder.flush()

    def start_span(self, *args, **kwargs):
        span = super().start_span(*args, **kwargs)
        if span.parent_id is None:
            span.context.trace_id = "1-%x-%s" % (
                int(time.time()),
                binascii.b2a_hex(os.urandom(12)).decode("ascii"),
            )
        span.context.span_id = generate_64bit_id_as_hex()
        return span

    def __enter__(self) -> "_XRayTracer":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.flush()



import traceback
from typing import Any, Type


def default_encoder(o: Any) -> Any:
    kind: Type[Any] = type(o)

    if issubclass(kind, BaseException):
        o: BaseException
        tb = o.__traceback__
        result = {
            "exception": o.__class__.__name__,
            "traceback": traceback.format_tb(tb),
            "message": str(o),
        }
        return result

    raise TypeError(
        f"Object of type {o.__class__.__name__} is not JSON serializable"
    )



from basictracer.span import BasicSpan
from opentracing.logs import ERROR_OBJECT

from xray_ot.settings import RecordingSettings
from xray_ot.translator.translator import preprocess_component_tags, Translator


DEFAULT_SETTINGS = RecordingSettings()


def test_preprocess_component_tags():
    tags_in = {
        "http.url": "https://example.com/",
        "http.method": "POST",
        "error": False,
        "hello": "",
        "hello.world": "",
        "abc": "",
    }
    tags_out = {"url": "https://example.com/", "method": "POST"}

    results = preprocess_component_tags("http", tags_in)
    assert results == tags_out


def test_translate_db_tags():
    tags_in = {
        "component": "db",
        "db.url": "postgres://127.0.0.1/db",
        "db.user": "postgres",
        "db.type": "postgres",
        "db.statement": "SELECT 1;",
    }
    tags_out = {
        "sql": {
            "url": tags_in["db.url"],
            "user": tags_in["db.user"],
            "database_type": tags_in["db.type"],
            "sanitized_query": tags_in["db.statement"],
        },
    }

    translator = Translator(DEFAULT_SETTINGS)
    results = translator.process_component(tags_in)

    assert results == tags_out


def test_translate_unknown_component_tags():
    tags_in = {
        "component": "graphql",
        "graphql.mutation": "createUser",
        "error": False,
    }
    tags_out = {
        "annotations": {"graphql_mutation": tags_in["graphql.mutation"],},
    }

    translator = Translator(DEFAULT_SETTINGS)
    results = translator.process_component(tags_in)

    assert results == tags_out


def test_translate_missing_component_tags():
    tags_in = {
        "error": False,
    }
    tags_out = {
        "annotations": {"_error": False,},
    }

    translator = Translator(DEFAULT_SETTINGS)
    results = translator.process_component(tags_in)

    assert results == tags_out


def test_translate_http_tags():
    tags_in = {
        "component": "http",
        "http.method": "POST",
        "http.url": "https://example.com/",
        "http.status_code": 405,
        "http.content_length": 5000,
        "http.other_repl": "abc",
    }
    tags_out = {
        "http": {
            "request": {"method": tags_in["http.method"], "url": tags_in["http.url"],},
            "response": {
                "status": tags_in["http.status_code"],
                "content_length": tags_in["http.content_length"],
                "other_repl": tags_in["http.other_repl"],
            },
        },
    }

    translator = Translator(DEFAULT_SETTINGS)
    results = translator.process_component(tags_in)

    assert results == tags_out


def test_translate_error_fatal_error():
    span = BasicSpan(tracer=None)

    tags_in = {
        "component": "http",
        "http.method": "POST",
        "error": True,
    }

    for tag_name, tag_value in tags_in.items():
        span.set_tag(tag_name, tag_value)

    span.log_kv({ERROR_OBJECT: ValueError("Expected an integer but got a string")})

    tags_out = {
        "fault": True,
        "http": {"request": {"method": "POST"}, "response": {}},
        "exception": {
            "message": "Expected an integer but got a string",
            "type": "ValueError",
        }
    }

    results = Translator.translate(settings=DEFAULT_SETTINGS, span=span)
    exception_id = results["exception"].pop("id")
    assert len(exception_id) == 16
    assert results == tags_out


def test_translate_error_handled_error():
    span = BasicSpan(tracer=None)

    tags_in = {
        "component": "http",
        "http.method": "POST",
        "error": True,
    }

    for tag_name, tag_value in tags_in.items():
        span.set_tag(tag_name, tag_value)

    tags_out = {
        "error": True,
        "http": {"request": {"method": "POST"}, "response": {}},
    }

    results = Translator.translate(settings=DEFAULT_SETTINGS, span=span)
    assert results == tags_out



from dataclasses import dataclass, field
from typing import Dict


def default_field(obj):
    return field(default_factory=lambda: obj)


@dataclass(frozen=True)
class RecordingSettings:
    db_annotations: Dict[str, str] = default_field(
        {
            "statement": "sanitized_query",
            "type": "database_type",
            # Implicit Associations:
            # db.user -> user
        }
    )

    http_request_annotations: Dict[str, str] = default_field(
        {
            "method": "method",
            "url": "url",
            "client_ip": "client_ip",
            "user_agent": "user_agent",
        }
    )

    http_response_annotations: Dict[str, str] = default_field({"status_code": "status"})



"""
AWS X-Ray backed implementation of the python OpenTracing API.

https://github.com/opentracing/basictracer-python

See the API definition for comments.
"""

import atexit
import sys
import threading
import time
import traceback
import warnings
from typing import List

from basictracer.recorder import SpanRecorder
from basictracer.span import BasicSpan

from . import connection as conn
from . import constants
from .settings import RecordingSettings
from .translator import Translator


class Recorder(SpanRecorder):
    """Recorder translates, buffers, and reports basictracer.BasicSpans.

    These reports are sent to an X-Ray daemon at the provided address.

    For parameter semantics, see Tracer() documentation; Recorder() respects
    component_name, collector_host, collector_port, tags, max_span_records,
    periodic_flush_seconds, verbosity
    """

    def __init__(
        self,
        component_name=None,
        collector_host="127.0.0.1",
        collector_port=2000,
        settings: RecordingSettings = None,
        max_span_records=constants.DEFAULT_MAX_SPAN_RECORDS,
        periodic_flush_seconds=constants.FLUSH_PERIOD_SECS,
        verbosity=0,
    ):
        self.verbosity = verbosity
        self._collector_address = (collector_host, collector_port)

        self.component_name = component_name if component_name is None else sys.argv[0]
        self.settings = settings if settings is not None else RecordingSettings()

        self._mutex = threading.Lock()
        self._span_records = []
        self._max_span_records = max_span_records

        self._disabled_runtime = False
        atexit.register(self.shutdown)

        self._periodic_flush_seconds = periodic_flush_seconds
        # _flush_connection and _flush_thread are created lazily since some
        # Python environments (e.g., Tornado) fork() initially and mess up the
        # reporting machinery up otherwise.
        self._flush_connection = None
        self._flush_thread = None
        if self._periodic_flush_seconds <= 0:
            warnings.warn(
                f"Runtime(periodic_flush_seconds={self._periodic_flush_seconds}) "
                f"means we will never flush to xray daemon unless explicitly requested."
            )

    def _maybe_init_flush_thread(self) -> None:
        """Start a periodic flush mechanism for this recorder if:

        1. periodic_flush_seconds > 0, and 
        2. self._flush_thread is None, indicating that we have not yet
           initialized the background flush thread.

        We do these things lazily because things like `tornado` break if the
        background flush thread starts before `fork()` calls happen.
        """
        if (self._periodic_flush_seconds > 0) and (self._flush_thread is None):
            self._flush_connection = conn.XRayConnection(self._collector_address)
            self._flush_thread = threading.Thread(
                target=self._flush_periodically, name=constants.FLUSH_THREAD_NAME
            )
            self._flush_thread.daemon = True
            self._flush_thread.start()

    def _fine(self, fmt, args) -> None:
        if self.verbosity >= 1:
            print("[X-Ray Tracer]:", (fmt % args))

    def _finest(self, fmt, args) -> None:
        if self.verbosity >= 2:
            print("[X-Ray Tracer]:", (fmt % args))

    def record_span(self, span) -> None:
        """Per BasicSpan.record_span, safely add a span to the buffer.

        Will drop a previously-added span if the limit has been reached.
        """
        if self._disabled_runtime:
            return

        # Lazy-init the flush loop (if need be).
        self._maybe_init_flush_thread()

        # Checking the len() here *could* result in a span getting dropped that
        # might have fit if a report started before the append(). This would only
        # happen if the client lib was being saturated anyway (and likely
        # dropping spans). But on the plus side, having the check here avoids
        # doing a span conversion when the span will just be dropped while also
        # keeping the lock scope minimized.
        with self._mutex:
            if len(self._span_records) >= self._max_span_records:
                return

        segment = {
            "trace_id": span.context.trace_id,
            "id": span.context.span_id,
            "start_time": span.start_time,
            "end_time": span.start_time + span.duration,
            "name": span.operation_name,
        }

        if span.parent_id is not None:
            segment["parent_id"] = span.parent_id
            segment["type"] = "subsegment"
        if span.tags:
            segment.update(Translator.translate(settings=self.settings, span=span))

        if len(span.logs) > 0:
            logs = [
                {"timestamp": log.timestamp, "fields": log.key_values}
                for log in span.logs
            ]
            segment["metadata"] = {"logs": logs}

        with self._mutex:
            if len(self._span_records) < self._max_span_records:
                self._span_records.append(segment)

    def flush(self, connection=None) -> bool:
        """Immediately send unreported data to the daemon.

        Calling flush() will ensure that any current unreported data will be
        immediately sent to the host server.

        If connection is not specified, the report will sent to the server
        passed in to __init__.  Note that custom connections are currently used
        for unit testing against a mocked connection.

        Returns whether the data was successfully flushed.
        """
        if self._disabled_runtime:
            return False

        if connection is not None:
            return self._flush_worker(connection)
        else:
            self._maybe_init_flush_thread()
            return self._flush_worker(self._flush_connection)

    def shutdown(self, flush=True) -> bool:
        """Shutdown the Runtime's connection by (optionally) flushing the
        remaining logs and spans and then disabling the Runtime.

        Note: spans and logs will no longer be reported after shutdown is called.

        Returns whether the data was successfully flushed.
        """
        # Closing connection twice results in an error. Exit early
        # if runtime has already been disabled.
        if self._disabled_runtime:
            return False

        flushed = flush and self.flush()
        self._disabled_runtime = True
        return flushed

    def _flush_periodically(self) -> None:
        """Periodically send reports to the server.

        Runs in a dedicated daemon thread (self._flush_thread).
        """
        # Send data until we get disabled
        while not self._disabled_runtime:
            self._flush_worker(self._flush_connection)
            time.sleep(self._periodic_flush_seconds)

    def _flush_worker(self, connection) -> bool:
        """Use the given connection to transmit the current logs and spans as a
        report request."""
        if connection is None:
            return False

        report_request = self._construct_report_request()
        try:
            self._finest("Sending report to daemon: %s", (report_request,))
            connection.report(report_request)

            return True

        except Exception as e:
            self._fine(
                "Caught exception during report: %s, stack trace: %s",
                (e, traceback.format_exc()),
            )
            self._restore_spans(report_request)
            return False

    def _construct_report_request(self) -> List[BasicSpan]:
        """Construct a report request."""
        with self._mutex:
            records = self._span_records
            self._span_records = []
        return records

    def _restore_spans(self, span_records) -> None:
        """Called after a flush error to move records back into the buffer
        """
        if self._disabled_runtime:
            return

        with self._mutex:
            if len(self._span_records) >= self._max_span_records:
                return
            combined = span_records + self._span_records
            self._span_records = combined[-self._max_span_records :]



#!/usr/bin/env python

from setuptools import find_packages, setup

setup(
    name="xray-python-opentracing-fork",
    version="0.0.3",
    description="AWS X-Ray Python OpenTracing Implementation",
    long_description="",
    author="Jeremy Apthorp <nornagon@nornagon.net>",
    maintainer="KOCAK Mikail <hello@vanille.bid>",
    license="",
    install_requires=["basictracer>=3.1,<4.0", "opentracing>=2.3.0"],
    tests_require=["pytest"],
    classifiers=[
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    keywords=[
        "opentracing",
        "aws",
        "xray",
        "awsxray",
        "traceguide",
        "tracing",
        "microservices",
        "distributed",
    ],
    packages=find_packages(exclude=["docs*", "tests*", "sample*"]),
)



from typing import Callable, Dict, Union, Optional

from basictracer.span import BasicSpan, LogData
from opentracing.logs import ERROR_OBJECT

from ..rand import generate_64bit_id_as_hex
from ..settings import RecordingSettings
from . import components as component_processors
from .ot_types import T_TAGS, T_XRAY_TAGS

T_COMPONENT_PROCESSOR = Callable[[RecordingSettings, T_TAGS], T_XRAY_TAGS]


def preprocess_component_tags(component: str, tags: T_TAGS) -> T_TAGS:
    """
    Strip the prefixes from the tags.

    Example 1:
        >>> tags: T_TAGS = {"http.url": "https://example.com/", "http.method": "POST"}
        >>> preprocess_component_tags(tags)
        >>> -> {"url": "https://example.com/", "method": "POST"}

    Example 2: removing other component fields, such as error[.object]
        >>> tags: T_TAGS = {"http.method": "POST", "error": True, "error.object": ...}
        >>> preprocess_component_tags(tags)
        >>> -> {"method": "POST"}

    """
    translation = {}
    tag_prefix_len = len(component)
    tag_prefix_len_with_sep = tag_prefix_len + 1  # {component} + '.'

    for tag_name, value in tags.items():
        # Skip the tag if it doesn't match the component name
        if len(tag_name) <= tag_prefix_len_with_sep:
            continue

        tag_prefix, tag_name = (
            tag_name[:tag_prefix_len],
            tag_name[tag_prefix_len_with_sep:],
        )

        # Skip the tag if it doesn't match the component name
        if tag_prefix != component:
            continue

        translation[tag_name] = value

    return translation


class Translator:
    def __init__(self, settings: RecordingSettings):
        self.settings = settings

    @classmethod
    def translate(cls, *, settings: RecordingSettings, span: BasicSpan) -> T_XRAY_TAGS:
        translator = cls(settings)
        payload = translator.process_component(span.tags)
        payload.update(translator.process_errors(span))
        return payload

    def process_component(self, tags: T_TAGS) -> Dict[str, T_TAGS]:
        component: str = tags.pop("component", "")
        translator = None

        if component:
            tags = preprocess_component_tags(component, tags)
            translator = self.get_component_translator(component)

        if translator is not None:
            translation = translator(self.settings, tags)
        else:
            translation = {
                "annotations": {
                    f"{component}_{tag}": value for tag, value in tags.items()
                }
            }
        return translation

    @staticmethod
    def process_errors(span: BasicSpan) -> Union[Dict, T_XRAY_TAGS]:
        results = {}

        if not span.tags.get("error", False):
            return results

        log: LogData
        error_obj = None

        for log in span.logs:
            error_obj: Optional[Exception] = log.key_values.get(ERROR_OBJECT, None)

        if error_obj is not None:
            # 5XX
            exception = {
                "id": generate_64bit_id_as_hex(),
                "message": str(error_obj),
                "type": error_obj.__class__.__name__,
            }
            results["fault"] = True
            results["exception"] = exception
        else:
            # 4XX
            results["error"] = True

        return results

    @staticmethod
    def get_component_translator(component) -> T_COMPONENT_PROCESSOR:
        translator: T_COMPONENT_PROCESSOR = getattr(
            component_processors, component, None
        )
        return translator
